<?php

namespace IMATHUZH\Qfq\Core\Report;

use Exception;
use IMATHUZH\Qfq\Core\Helper\KeyValueStringParser;
use IMATHUZH\Qfq\Core\Store\Store;

/**
 * Class RestClient
 * @package IMATHUZH\Qfq\Core\Report
 */
class RestClient {

    /**
     * @var Store
     */
    private $store = null;

    /**
     * RestClient constructor.
     * @throws \CodeException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public function __construct() {
        $this->store = Store::getInstance();
    }

    /**
     * @param string $str
     * @return string
     * @throws \CodeException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public function process($str) {
        $recvBuffer = '';
        $recv = array();

        $param = $this->parseArgument($str);

//        if (isset($param[TOKEN_L_SSL])) {
//            $options['ssl'] = json_decode($param[TOKEN_L_SSL], true);
//        }

        // Send request
        try {
            list($http_status, $recvBuffer) = self::callApiCurl(strtoupper($param[TOKEN_L_METHOD]), $param[TOKEN_REST_CLIENT], $data = $param[TOKEN_L_CONTENT] ?? '', $param[TOKEN_L_HEADER], $param[TOKEN_L_TIMEOUT]);
            $recv[HTTP_STATUS] = $http_status;
            if ($http_status >= 300) {
                $recv[ERROR_MESSAGE] = ' Api error: ' . $param[TOKEN_REST_CLIENT] . '  HTTP code: ' . $http_status . '  Message: ' . print_r(json_decode($recvBuffer ?? '', true), true);
            }
            $rcvBufferDecoded = json_decode($recvBuffer, true);
            if ($rcvBufferDecoded !== Null) {
                array_walk_recursive($rcvBufferDecoded, function ($value, $key) use (&$recv) {
                    $recv[$key] = $value;
                });
            }
        } catch (\Exception $e) {
            $recv[HTTP_STATUS] = $e->getCode();
            $recv[ERROR_MESSAGE] = $e->getMessage();
        }

        // Copy new values to STORE_CLIENT
        $this->store::setStore($recv, STORE_CLIENT, true);
        return $recvBuffer;
    }

    /**
     * Parses $str, fill some defaults and returns an array with given arguments.
     *
     * @param string $str
     * @return array
     * @throws \UserFormException
     * @throws \UserReportException
     */
    private function parseArgument($str) {
        // "n:http://antmedia-dev.math.uzh.ch/WebRTCAppEE/rest/v2/broadcasts/create|
        //  content:{'streamId' => "ASDKLJfdlajfhdkhH"}|method:POST|header:Content-type: application/json\r\n"

        // Split string
        $param = KeyValueStringParser::parse($str, PARAM_TOKEN_DELIMITER, PARAM_DELIMITER);
        if (empty($param[TOKEN_REST_CLIENT])) {
            throw new \UserReportException("Missing RestClient target", ERROR_MISSING_VALUE);
        }

        $param[TOKEN_L_CONTENT] = trim($param[TOKEN_L_CONTENT]);
        $param[TOKEN_L_HEADER] = trim($param[TOKEN_L_HEADER]);

        if (empty($param[TOKEN_L_METHOD])) {
            $param[TOKEN_L_METHOD] = 'GET';
        }

        // Default Timeout
        if (empty($param[TOKEN_L_TIMEOUT])) {
            $param[TOKEN_L_TIMEOUT] = 5;
        }

        // split header at "\r\n" and "\n". $header must be an array of single headers!!
        if (!empty($param[TOKEN_L_HEADER])) {
            $header = str_replace("\r\n", '$$SEP$$', $param[TOKEN_L_HEADER]);
            $header = str_replace("\n", '$$SEP$$', $header);
            $header = explode('$$SEP$$', $header);
        } else {
            $header = [];
        }

        // If 'Host' is missing in header: define - useful for Firewall/ Proxy
        // CR: if a header 'host' is given, REST calls fails always.
//        if (strpos($param[TOKEN_L_HEADER], 'host:') === false) {
//            $urlParts = parse_url($param[TOKEN_REST_CLIENT]);
//            $header[] = 'host: ' . $urlParts['host'];
//        }

        $param[TOKEN_L_HEADER] = $header;
        return $param;
    }

    /**
     * @param string $method
     * @param string $url
     * @param string $data
     * @param array $header
     * @param int $timeout
     * @return array
     * @throws Exception
     */
    public static function callApiCurl(string $method, string $url, string $data = '', array $header = [], int $timeout = 5) {

        // Header: Set content-type if not set
        if (strpos(join('', $header), 'content-type:') === false) {
            $mime = (($data[0] ?? '') == '{') ? 'application/json' : 'text/plain';
            $header[] = 'content-type: ' . $mime . '; charset=utf-8';
        }

        // Header: Set connection if not set
        if (strpos(join('', $header), 'connection:') === false) {
            $header[] = 'connection: close';
        }

        $ch = curl_init();
        $curlConfig = array(
            CURLOPT_RETURNTRANSFER => true,
            // CURLINFO_HEADER_OUT => $debug
        );
        switch ($method) {
            case "POST":
                $curlConfig[CURLOPT_POST] = true;
                if (!empty($data)) {
                    $dataJson = $data;
                    $curlConfig[CURLOPT_POSTFIELDS] = $dataJson;
                    $header[] = 'Content-Length: ' . strlen($dataJson);
                }
                break;
            case "PUT":
                $curlConfig[CURLOPT_CUSTOMREQUEST] = 'PUT';
                if (!empty($data)) {
                    $dataJson = $data;
                    $curlConfig[CURLOPT_POSTFIELDS] = $dataJson;
                    $header[] = 'Content-Length: ' . strlen($dataJson);
                }
                break;
            case "DELETE":
                $curlConfig[CURLOPT_CUSTOMREQUEST] = 'DELETE';
                if (!empty($data)) {
                    $url = sprintf("%s?%s", $url, http_build_query(json_decode($data, true)));
                }
                break;
            default:
                if (!empty($data)) {
                    $url = sprintf("%s?%s", $url, http_build_query(json_decode($data, true)));
                }
        }
        $curlConfig[CURLOPT_URL] = $url;
        $curlConfig[CURLOPT_TIMEOUT] = $timeout;
        curl_setopt_array($ch, $curlConfig);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $header);

        // send request
        $output = curl_exec($ch);
        if ($output === false) {
            throw new Exception(curl_error($ch), curl_errno($ch));
        }
        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);
        return [$httpCode, $output];
    }
}
